/*
 * BLLib.Net.Packet
 * Andy Tidball
 * 
 * Project: Black Lightning General Library
 * Copyright: GNU General Public License
 */

using System;

namespace BLLib.Net {
	/// <summary>
	/// Abstract base class for packets to be sent or received through a PacketConnection.
	/// </summary>
	[Serializable]
	public abstract class Packet {
		// member data
		private string   _Data;         // the data string contained in the packet
		private DateTime _CreationTime; // the date and time that the packet was created

		/// <summary>
		/// Creates a new Packet.
		/// </summary>
		/// <param name="Data">The data string contained in the packet.</param>
		public Packet(string Data) {
			_Data         = Data;
			_CreationTime = DateTime.Now;
		}

		/// <summary>
		/// Copy constructor.
		/// </summary>
		/// <param name="P">The packet to create a copy of.</param>
		public Packet(Packet P) {
			_Data         = P._Data;
			_CreationTime = P._CreationTime;
		}

		/// <summary>
		/// Gets the data string that the packet was created with.
		/// </summary>
		public string RawInput {
			get {
				return _Data;
			}
		}

		/// <summary>
		/// The date and time that this packet was created.
		/// </summary>
		public DateTime CreationTime {
			get {
				return _CreationTime;
			}
		}

		/// <summary>
		/// Constructs an output raw data string from the current state of whatever constructs are used by the derived implementation.
		/// </summary>
		/// <returns>A data string that reflects the current state of the derived packet.</returns>
		public abstract string CreateOutput();
	}

	/// <summary>
	/// A packet which doesn't break the data string down into any more detail.
	/// </summary>
	[Serializable]
	public class RawPacket : Packet {
		// member data
		private string _Data; // the packet's data

		/// <summary>
		/// Creates a new RawPacket.
		/// </summary>
		/// <param name="Data">The data to contain in the packet.</param>
		public RawPacket(string Data)
			: base(Data) {
            _Data = this.RawInput;
		}

        /// <summary>
        /// Copy constructor.
        /// </summary>
        /// <param name="P">The packet to create a copy of.</param>
        public RawPacket(RawPacket P)
            : base(P) {
            _Data = P._Data;
        }

		/// <summary>
		/// The data contained in the packet.
		/// </summary>
		public string Data {
			get {
				return _Data;
			}
			set {
				_Data = value;
			}
		}

		/// <summary>
		/// Retrieves the packet's data.
		/// </summary>
		/// <returns>The packet's data string.</returns>
		public override string CreateOutput() {
			return _Data;
		}
	}
}
