/*
 * BlackLightning.Wizard
 * Andy Tidball
 * 
 * Project: Black Lightning
 * Copyright: GNU General Public License
 */

using System;
using System.Text.RegularExpressions;
using BlackLightning.PluginLib.Modules;
using BLLib.Net;

namespace BlackLightning {
	/// <summary>
	/// Represents a connection to the Wizard front end program.
	/// </summary>
	internal class Wizard : FrontEnd {
		/// <summary>
		/// Creates a new Wizard connection.
		/// </summary>
		public Wizard()
			: base("\n", true, "\n", "\n") {
		}

		/// <summary>
		/// Gets the name of the front end's executable file.
		/// </summary>
		public override string ExecutableFilename {
			get {
				return "Wizard.exe";
			}
		}

		/// <summary>
		/// Sends a Packet down the connection.
		/// </summary>
		/// <param name="Packet">The packet to send down the connection.</param>
		public override void SendPacket(Packet Packet) {
			// create appropriate contextual information
			base.SendPacket(Packet, new BLPacketContext(Protocol.GSL));
		}

		/// <summary>
		/// Creates a Packet from the front end and headed for either the game server or Black Lightning.
		/// </summary>
		/// <param name="Data">The data to initialize the packet with.</param>
		protected override Packet CreatePacket(string Data) {
			if (Data == "GOOD" + Environment.NewLine) {
				// this is the validation achknowledgement packet, it's not a command
				return new RawBLPacket(Program.FrontEnd, Program.GameServer, Data);
			} else {
				// anything else sent by the Wizard (that I know about) is a command
				return new CommandPacket(Program.FrontEnd, (Data.StartsWith("@") ? Program.BlackLightning : Program.GameServer), Data, new BLPacketContext(Protocol.GSL));
			}
		}

		/// <summary>
		/// Handshakes with the game server to login.
		/// </summary>
		/// <param name="Server">The new connection to the game server to handshake with.</param>
		public override void Handshake(Connection Server, bool FirstConnection) {
			string Received;
			if (FirstConnection) {
				// if this is the first connection, receive Wizard's login commands
				Regex LoginCmds = new Regex("([0-9a-f]*)\n(.*)\n");
				Received = base.Connection.Receive();
				while (!LoginCmds.IsMatch(Received)) {
					Received += base.Connection.Receive();
				}

				// store the identification
				Match M = LoginCmds.Match(Received);
				base.Identification = M.Groups[2].Captures[0].Value;

				// send the commands on to the server
				Server.Send(Received);
			} else {
				// if we're connecting for a second time, send our new key and stored identification
				Server.Send(base.SalFile.Key + "\n" + base.Identification + "\n");
			}

			// wait until we receive our settings info, and send that back to StormFront
			Regex SettingsInfo = new Regex(".*\n.*\r\n");
			Received = Server.Receive();
			while (!SettingsInfo.IsMatch(Received)) {
				Received += Server.Receive();
			}
			base.Connection.Send(Received);
/*
			// if we're reconnecting then we have to send the GOOD packet because the Wizard won't
			if (!FirstConnection) {
				Server.Send("GOOD\n");
			}

			// wait for the blank line back
			Regex BlankLine = new Regex("<c>\n");
			Received = base.Connection.Receive();
			while (!BlankLine.IsMatch(Received)) {
				Received += base.Connection.Receive();
			}
*/
		}
	}
}
